#!/vol/PACK/perl5.004_04/bin/perl -w
#
# ConnDB.pm
#
# Routines for manipulating the connections database
#
# Andrew.Findlay@brunel.ac.uk
# August 1999
#
# @(#) ConnDB.pm 1.49@(#)

########################################################################
# Boilerplate stuff to make this into a perl module
########################################################################
package ConnDB;

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
	$errstr
	Connect
	FetchDBDate
	AddUser
	AddServiceRequest
	AddServiceRequestItem
	AddCardDetails
	AddConnection
	DeleteConnection
	FetchConnection
	FetchConnectionList
	FetchFullName
	FetchSiteList
	FetchActiveUserList
	AddBuildingToDB
	FetchBuildingList
	FetchBuildingName
	HasBlocks
	FetchOutlets
	FetchSemesterList
	FetchRenewalSemesterList
	FetchSemester
	FetchRequests
	FetchBillingDetails
	FetchPendingConnections
	FetchRequestItem
	FetchRequestItemsByRequestID
	HallsRequestExists
	FetchPendingCardPayments
	FetchPendingCashPayments
	FetchLatePayments
	FetchExpiringConnections
	AddDeviceToDB
	FetchDeviceList
	ProcessPrePayment
	ValidateInputValue
	MarkExpiredConnections
	CancelRequestItem
	FetchPermissionsList
	FetchPrivsList
	SetPermissions
	Permitted
	LoadOutletData
	KnownOutletNumber
	ApplyForExtension
	GetCounts
);
########################################################################
########################################################################

use strict;
use DBI;
use ConnLog;


# How far in advance should connections appear on the list of jobs?
my ($CONST_days_warning_of_connections) = 7;

# How large may comments be?
my ($CONST_max_comment_length) = 64;

# What is the largest state number?
my ($CONST_max_state_num) = 7;

# Error strings stored here - this one is exported
$ConnDB::errstr = undef;

# Database handle 
my ($dbh);

########################################################################
#
# Connect to the database. Return zero if OK, error string if not.
# The parameter is the name of the file where database connection
# info is stored
#
########################################################################
sub Connect {
	my ($database_param_file) = @_;

	my ($DatabaseID, $DatabaseUser, $DatabasePW);

	# File contains lines like this:
	# DatabaseID: DBI:mysql:database=connections;host=freya.brunel.ac.uk
	# DatabaseUser: andrew
	# DatabasePW: 809eB_qK

	unless (open PARAMFILE, $database_param_file) {
		return $ConnDB::errstr = "Cannot read database parameter file";
	}

	while (<PARAMFILE>) {
		# Clean up the lines
		chomp;

		# Ignore blanks and comments
		next if /^\s*$/;
		next if /^\s*#/;

		# Split the line
		my ($left, $right) = split /\s+/, $_, 2;
		$DatabaseID = $right if ($left eq "DatabaseID:");
		$DatabaseUser = $right if ($left eq "DatabaseUser:");
		$DatabasePW = $right if ($left eq "DatabasePW:");
	}
	close PARAMFILE;

	if (!defined( $DatabaseID )) {
		$ConnDB::errstr = "DatabaseID not found in parameter file";
		return $ConnDB::errstr;
	}

	if (!defined( $DatabaseUser )) {
		$ConnDB::errstr = "DatabaseUser not found in parameter file";
		return $ConnDB::errstr;
	}

	if (!defined( $DatabasePW )) {
		$ConnDB::errstr = "DatabasePW not found in parameter file";
		return $ConnDB::errstr;
	}

	$dbh = DBI->connect(
		$DatabaseID, $DatabaseUser, $DatabasePW,
		{ RaiseError => 0, PrintError => 0, AutoCommit => 1 } );

	if (!defined($dbh)) {
		return $ConnDB::errstr = $DBI::errstr;
	} else {
		return 0;
	}
}

########################################################################
#
# Validate an input value by looking it up in the database
# It must be non-null and match at least one item in the specified
# table/column.
#
########################################################################
sub ValidateInputValue {
	my ($tablename, $colname, $value) = @_;

	if (!defined($value)) {
		return 0;
	}

	# We need at least one item in the column
	# We cannot use the normal cached-template here
	# as the execute methods cannot substitute for
	# table and column names.
	# Note the use of qq to allow substitution here.
	my ($sth_get) = $dbh->prepare(qq{
		SELECT
			COUNT(*)
		FROM $tablename
		WHERE $colname = ?
	});
	if (!defined($sth_get)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# How many are there?
	my ($number) = $dbh->selectrow_array($sth_get,undef,($value));
	if (!defined($number)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# No matching values found
	if ($number < 1) {
		return 0;
	}

	# OK - value is in table
	return 1;
}

########################################################################
#
# Get the database's idea of the current date and time
#
########################################################################
sub FetchDBDate {

	my ($sth_date) = $dbh->prepare_cached(q{
		SELECT NOW()
	});
	if (!defined($sth_date)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	my ($now) = $dbh->selectrow_array($sth_date);
	if (!defined($now)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# OK - return the date/time string
	return $now;
}

########################################################################
#
# Has the specified building got flats or blocks?
# Return non-zero if it has, zero if not.
# Returns zero for error as well - may not be clever...
#
########################################################################
sub HasBlocks {
	my ($building) = @_;

	# We need at least one item in the column
	my ($sth_get) = $dbh->prepare_cached(q{
		SELECT has_subunits
		FROM buildings
		WHERE building_id = ?
	});
	if (!defined($sth_get)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the result
	my ($result) = $dbh->selectrow_array($sth_get,undef,($building));
	if (!defined($result)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $result;
}

########################################################################
#
# Add an entry in the user table
# Returns non-zero on success, zero on fail
#
########################################################################
sub AddUser {
	my ($username) = @_;

	my ($user,$pass,$uid,$gid,$quota,$comment,$gcos,$dir,$shell) =
		getpwnam $username;
	# print "$user,$pass,$uid,$gid,$quota,$comment,$gcos,$dir,$shell\n";

	if (!defined($user)) {
		$ConnDB::errstr = "User '$username' is not known";
		return 0;
	}

	$comment =~ s/,.*//;
	chomp $comment;

	# Prepare and cache the insert statement
	my ($sth_adduser) = $dbh->prepare_cached(q{
		INSERT INTO user (username,fullname)
		VALUES (?,?)
	});
	if (!defined($sth_adduser)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	if (!$sth_adduser->execute( $user, $comment ))
	{
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	Log( "notice", "OK AddUser $username" );
	return 1;
}

########################################################################
#
# Add an entry in the service request table
# Returns request ID on success, zero on fail.
# Assumes that all data has been sanity-checked
#
########################################################################
sub AddServiceRequest {
	my (
		$username,
		$service_type,
		$charge,
		$billing_details_id,
		$comment
	) = @_;

	# Prepare and cache the insert statement
	my ($sth_addreq) = $dbh->prepare_cached(q{
		INSERT INTO service_request (
			service_request_id,
			username,
			request_date,
			service_type,
			charge,
			amount_paid,
			billing_details_id,
			comment
		)
		VALUES (
			NULL,		# service_request_id is auto-inc
			?,		# username
			NOW(),		# request_date is auto-set
			?,		# service_type
			?,		# charge
			0,		# nothing paid yet
			?,		# billing_details_id
			?		# comment
		)
	});
	if (!defined($sth_addreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Insert the data
	if (!$sth_addreq->execute(
		$username,
		$service_type,
		$charge,
		$billing_details_id,
		$comment
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Now find out what request_id was allocated to that record
	my ($req_id) = $dbh->selectrow_array(q{
		SELECT LAST_INSERT_ID()
	});

	Log( "notice", "OK AddServiceRequest ($req_id) $username, $service_type, $charge, $billing_details_id, $comment" );

	return $req_id;
}

########################################################################
#
# Add an entry in the service request item table
# Returns item ID on success, zero on fail.
# Assumes that all data has been sanity-checked
#
########################################################################
sub AddServiceRequestItem {
	my (
		$service_request_id,
		$building_id,
		$block_number,
		$room_number,
		$outlet_number,
		$port_type,
		$start_date,
		$end_date,
		$service_state_id,
		$charge,
		$comment
	) = @_;

	# Prepare and cache the insert statement
	my ($sth_addreq) = $dbh->prepare_cached(q{
		INSERT INTO service_request_item (
			service_request_id,
			building_id,
			block_number,
			room_number,
			outlet_number,
			port_type,
			start_date,
			end_date,
			service_state_id,
			charge,
			comment
		)
		VALUES (
			?,?,?,?,?,?,?,?,
			?,		# service_state_id depends on billing
			?,?
		)
	});
	if (!defined($sth_addreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Insert the data
	if (!$sth_addreq->execute(
		$service_request_id,
		$building_id,
		$block_number,
		$room_number,
		$outlet_number,
		$port_type,
		$start_date,
		$end_date,
		$service_state_id,
		$charge,
		$comment
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Now find out what request_id was allocated to that record
	my ($req_id) = $dbh->selectrow_array(q{
		SELECT LAST_INSERT_ID()
	});

	Log( "notice", "OK AddServiceRequestItem ($req_id) $service_request_id, $building_id, $block_number, $room_number, $outlet_number, $port_type, $start_date, $end_date, $service_state_id, $charge, $comment" );

	return $req_id;
}

########################################################################
#
# Add Credit/Debit Card payment details
# Returns billing_details_id on success, zero on failure
#
########################################################################
sub AddCardDetails
{
	my (
		$card_account_name,
		$card_number,
		$card_expiry,
		$card_issue
	) = @_;

	# Prepare and cache the insert statement
	my ($sth_addcard) = $dbh->prepare_cached(q{
		INSERT INTO billing_details (
			billing_details_id,
			card_account_name,
			card_number,
			card_expiry,
			card_issue,
			internal_charge_code
		)
		VALUES (
			NULL,?,?,?,?,""
		)
	});
	if (!defined($sth_addcard)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Insert the data
	if (!$sth_addcard->execute(
		$card_account_name,
		$card_number,
		$card_expiry,
		$card_issue
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Now find out what billing_details_id was allocated to that record
	my ($item_id) = $dbh->selectrow_array(q{
		SELECT LAST_INSERT_ID()
	});

	Log( "notice", "OK AddCardDetails $item_id" );

	return $item_id;
}

########################################################################
#
# Record application to extend a connection
#
########################################################################
sub ApplyForExtension {
	my (
		$service_request_id,
		$item_id,
		$sem_end_date,
		$sem_fee,
		$bill_id
	) = @_;

	# Prepare SQL for updating service_request
	my ($sth_update_rq) = $dbh->prepare_cached(q{
		UPDATE service_request
		SET
			charge = ?,
			amount_paid = 0,
			billing_details_id = ?,
			request_date = NOW(),
			paid_date = "0000-00-00"
		WHERE service_request_id = ?
	});
	if (!defined($sth_update_rq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the update
	if (!$sth_update_rq->execute(
		$sem_fee,
		$bill_id,
		$service_request_id
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Prepare SQL for updating service_request_item
	my ($sth_update_rqi) = $dbh->prepare_cached(q{
		UPDATE service_request_item
		SET
			renew_to_date = ?,
			service_state_id = 7,
			charge = ?
		WHERE service_request_item_id = ?
	});
	if (!defined($sth_update_rqi)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the update
	if (!$sth_update_rqi->execute(
		$sem_end_date,
		$sem_fee,
		$item_id
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# All OK
	return 1;
}


########################################################################
#
# Add Connection details
# Returns non-zero on success, zero on failure
#
########################################################################
sub AddConnection
{
	my (
		$building_id,
		$block_number,
		$room_number,
		$outlet_number,
		$device_id,
		$device_port,
		$service_request_item_id,
		$connected_by
	) = @_;

	# Sanity check: is this outlet already connected?
	my ($sth_c1) = $dbh->prepare_cached(q{
		SELECT COUNT(*)
		FROM connection
		WHERE building_id = ?
		AND block_number = ?
		AND outlet_number = ?
		AND date_disconnected = '0000-00-00 00:00:00'
	});
	if (!defined($sth_c1)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# How many are there?
	my ($existing_connections) = $dbh->selectrow_array(
			$sth_c1,undef,($building_id, $block_number, $outlet_number));
	if (!defined($existing_connections)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	if ($existing_connections > 0) {
		# Cannot connect one outlet to two devices!
		my ($place) = FetchBuildingName( $building_id );
		$place .= " block/flat $block_number" if ($block_number ne "");

		$ConnDB::errstr = "Outlet $outlet_number in $place is already connected!";
		return 0;
	}


	# Sanity check: is this device port already connected?
	my ($sth_c2) = $dbh->prepare_cached(q{
		SELECT COUNT(*)
		FROM connection
		WHERE device_id = ?
		AND device_port = ?
		AND date_disconnected = '0000-00-00 00:00:00'
	});
	if (!defined($sth_c2)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# How many are there?
	my ($existing_portconns) = $dbh->selectrow_array(
			$sth_c2,undef,($device_id, $device_port));
	if (!defined($existing_portconns)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	if ($existing_portconns > 0) {
		# Cannot connect one port to two outlets
		my ($place) = FetchBuildingName( $building_id );
		$place .= " block/flat $block_number" if ($block_number ne "");

		my ($device_name) = FetchDeviceName( $device_id );
		$ConnDB::errstr = "Port $device_port of $device_name in $place is already in use!";
		return 0;
	}

	# Prepare and cache the insert statement
	my ($sth_addconn) = $dbh->prepare_cached(q{
		INSERT INTO connection (
			building_id,
			block_number,
			room_number,
			outlet_number,
			device_id,
			device_port,
			service_request_item_id,
			date_connected,
			date_disconnected,
			connected_by,
			disconnected_by
		)
		VALUES (
			?,?,?,?,?,?,?,NOW(),0,?,''
		)
	});
	if (!defined($sth_addconn)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Insert the data
	if (!$sth_addconn->execute(
		$building_id,
		$block_number,
		$room_number,
		$outlet_number,
		$device_id,
		$device_port,
		$service_request_item_id,
		$connected_by
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# We now record the fact that the service_request_item is connected
	my ($sth_update) = $dbh->prepare_cached(q{
		UPDATE service_request_item
		SET service_state_id = '3'
		WHERE service_request_item_id = ?
	});
	if (!defined($sth_update)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}
	if (!$sth_update->execute( $service_request_item_id )) {
		# Failed to update...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Find out the ID allocated to this connection
	my ($connection_id) = $dbh->selectrow_array(q{
		SELECT LAST_INSERT_ID()
	});

	Log( "notice", "OK AddConnection ($connection_id) $building_id, $block_number, $room_number, $outlet_number, $device_id, $device_port, $service_request_item_id, $connected_by" );

	return 1;
}

########################################################################
#
# DeleteConnection($req_item_id, $whodidit)
# - delete connection from database
#
########################################################################

sub DeleteConnection {
	my ($req_item_id, $whodidit) = @_;

	# Prepare statement for connection table update
	my ($sth_delconn) = $dbh->prepare_cached(q{
		UPDATE connection
		SET date_disconnected = NOW(), disconnected_by = ?
		WHERE service_request_item_id = ?
	});
	if (!defined($sth_delconn)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}
	if (!$sth_delconn->execute( $whodidit, $req_item_id )) {
		# Failed to update...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Now for the service_request_item table
	my ($sth_delitem) = $dbh->prepare_cached(q{
		UPDATE service_request_item
		SET service_state_id = 5
		WHERE service_request_item_id = ?
	});
	if (!defined($sth_delitem)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}
	if (!$sth_delitem->execute( $req_item_id )) {
		# Failed to update...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	Log( "notice", "OK DeleteConnection $req_item_id, $whodidit" );

	# All OK
	return 1;
}


########################################################################
#
# Fetch details of a connection
# Returns array on success, zero on failure
#
# usage: FetchConnection <connection_id> <service_request_item_id>
# Either param may be given as "%"
#
########################################################################
sub FetchConnection
{
	my (
		$connection_id, $service_request_item_id
	) = @_;

	# Prepare and cache the select statement
	my ($sth_fetchreq) = $dbh->prepare_cached(q{
		SELECT DISTINCT
			connection.building_id,
			buildings.building_name,
			connection.block_number,
			connection.room_number,
			connection.outlet_number,
			connection.device_id,
			devices.device_name,
			connection.device_port,
			connection.service_request_item_id,
			connection.date_connected,
			connection.date_disconnected,
			connection.connected_by,
			connection.disconnected_by,
			service_request_item.service_state_id,
			service_states.service_state_description,
			service_request_item.comment
		FROM connection,service_request_item,buildings,service_states,devices
		WHERE connection.connection_id LIKE ?
		AND connection.service_request_item_id LIKE ?
		AND service_request_item.service_request_item_id = connection.service_request_item_id
		AND service_request_item.building_id=buildings.building_id
		AND service_request_item.service_state_id=service_states.service_state_id
		AND connection.device_id=devices.device_id
	});
	if (!defined($sth_fetchreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetchreq->execute(
		$connection_id,
		$service_request_item_id
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth_fetchreq->fetchall_arrayref;
	if ($sth_fetchreq->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}


	# There should only be one set of item details under this ID
	if (defined($requests->[1]) || !defined($requests->[0])) {
		$ConnDB::errstr =
			"Internal error: wrong number of connections for ID $connection_id";
		return 0;
	}

	# We want to return an array, not a reference:
	return @{$requests->[0]};
}


########################################################################
#
# AddDeviceToDB( $device_name, $building_id, $block, $device_type)
#
# Returns 0 for fail, 1 for success
########################################################################
sub AddDeviceToDB {
	my ( $device_name, $building_id, $block, $device_type) = @_;
	

	# Prepare and cache the insert statement
	my ($sth_adddev) = $dbh->prepare_cached(q{
		INSERT INTO devices (
			device_id,
			device_name,
			building_id,
			block_number,
			device_type
		)
		VALUES (NULL,?,?,?,?)
	});
	if (!defined($sth_adddev)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	if (!$sth_adddev->execute( $device_name, $building_id, $block, $device_type))
	{
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Find out the ID allocated to this device
	my ($device_id) = $dbh->selectrow_array(q{
		SELECT LAST_INSERT_ID()
	});

	Log( "notice", "OK AddDeviceToDB ($device_id) $device_name, $building_id, $block, $device_type" );

	# OK
	return 1;
}


########################################################################
#
# AddBuildingToDB( $building_code, $building_name, $building_type,
#				$has_blocks, $site_id))
#
# Returns 0 for fail, 1 for success
########################################################################
sub AddBuildingToDB {
	my ( $building_code, $building_name, $building_type,
		$has_blocks, $site_id) = @_;
	

	# Prepare and cache the insert statement
	my ($sth_addbldg) = $dbh->prepare_cached(q{
		INSERT INTO buildings (
			building_id,
			building_code,
			building_name,
			building_type,
			has_subunits,
			site_id
		)
		VALUES (NULL,?,?,?,?,?)
	});
	if (!defined($sth_addbldg)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	if (!$sth_addbldg->execute(
		$building_code, $building_name, $building_type,
		$has_blocks, $site_id ))
	{
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Find out the ID allocated to this building
	my ($building_id) = $dbh->selectrow_array(q{
		SELECT LAST_INSERT_ID()
	});

	Log( "notice", "OK AddBuildingToDB ($building_id) $building_code, $building_name, $building_type, $has_blocks, $site_id" );

	# OK
	return 1;
}

########################################################################
#
# Fetch name of a building
# Returns string on success, zero on failure
#
########################################################################
sub FetchBuildingName {
	my (
		$building_id
	) = @_;

	# Prepare and cache the select statement
	my ($sth_fetchreq) = $dbh->prepare_cached(q{
		SELECT
			building_name
		FROM buildings
		WHERE building_id = ?
	});
	if (!defined($sth_fetchreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Get the data
	my ($fullname) = $dbh->selectrow_array($sth_fetchreq,undef,($building_id));
	if (!defined($fullname)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $fullname;
}


########################################################################
#
# Fetch list of outlets given building_id and block_number
# Optionally limit by room number
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchOutlets
{
	my ($building_id, $block_number, $room_spec) = @_;

	# Prepare and cache the select statement
	my ($sth_fetch) = $dbh->prepare_cached(q{
		SELECT
			room_number,
			outlet_number
		FROM outlets
		WHERE building_id = ?
		AND block_number = ?
		AND room_number LIKE ?
		ORDER BY room_number, outlet_number
	});
	if (!defined($sth_fetch)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetch->execute($building_id, $block_number, $room_spec) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($outlets) = $sth_fetch->fetchall_arrayref;
	if ($sth_fetch->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $outlets;
}

########################################################################
#
# Fetch name of a device
# Returns string on success, zero on failure
#
########################################################################
sub FetchDeviceName {
	my (
		$building_id
	) = @_;

	# Prepare and cache the select statement
	my ($sth_fetchreq) = $dbh->prepare_cached(q{
		SELECT
			device_name
		FROM devices
		WHERE device_id = ?
	});
	if (!defined($sth_fetchreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Get the data
	my ($name) = $dbh->selectrow_array($sth_fetchreq,undef,($building_id));
	if (!defined($name)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $name;
}

########################################################################
#
# Fetch full name for a username
# Returns string on success, zero on failure
# Uses database for preference, but if second param is zero will try
# getpwnam if necessary.
# If second param is non-zero will only use database.
# Note that we dont just use getpwnam here because we may need to display
# details about defunct users that remain in the connections database
#
########################################################################
sub FetchFullName
{
	my (
		$username, $db_only
	) = @_;

	# Prepare and cache the select statement
	my ($sth_fetchreq) = $dbh->prepare_cached(q{
		SELECT
			fullname
		FROM user
		WHERE username = ?
	});
	if (!defined($sth_fetchreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Get the data
	my ($fullname) = $dbh->selectrow_array($sth_fetchreq,undef,($username));
	if (!defined($fullname)) {
		# Failed...
		if (defined($DBI::errstr)) {
			# Database problem: serious
			$ConnDB::errstr = $DBI::errstr;
			return 0;
		}
		else {
			# User not in database
			if ($db_only) {
				$ConnDB::errstr = "User '$username' is not known";
				return 0;
			}

			# Try the password file
			my ($user,$pass,$uid,$gid,$quota, $comment,$gcos,$dir,$shell) =
				getpwnam($username);

			if (!defined($user)) {
				$ConnDB::errstr = "User '$username' is not known";
				return 0;
			}

			$comment =~ s/,.*//;
			chomp $comment;
			$fullname = $comment;
		}
	}

	return $fullname;
}

########################################################################
# Fetch list of active users
#	FetchActiveUserList
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchActiveUserList
{
	# Prepare and cache the select statement
	# We want all users who have a connection now or will soon have one
	#
	my ($sth) = $dbh->prepare_cached(q{
		SELECT DISTINCT service_request.username, user.fullname
		FROM service_request,service_request_item,user
		WHERE service_request.service_request_id=service_request_item.service_request_id
		AND service_request.username=user.username
		AND service_request_item.service_state_id IN (1,2,3,7)
		ORDER BY service_request.username
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	#
	if (!$sth->execute() ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($users) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $users;
}

########################################################################
# Fetch list of devices available in a building/block:
#	FetchDeviceList <building_id> <block>
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchDeviceList
{
	my (
		$building_id, $block_number
	) = @_;

	# Prepare and cache the select statement
	#
	# NOTE: we are ignoring block number as some buildings
	#       have devices that can serve multiple blocks
	#
	#       Thus removed: 'AND block_number = ?' from the SQL
	#
	my ($sth) = $dbh->prepare_cached(q{
		SELECT
			device_id,
			device_name,
			device_type
		FROM devices
		WHERE building_id = ?
		ORDER BY device_name
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	# NOTE: removed $block_number from this (see above)
	#
	if (!$sth->execute(
		$building_id
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($devices) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $devices;
}


########################################################################
#
# Fetch list of buildings:
#	FetchBuildingList <typecode>
# If typecode is "%" then returns all buildings, else just those
# whose building_type matches the param.
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchBuildingList
{
	my (
		$typecode
	) = @_;

	# Prepare and cache the select statement
	my ($sth_fetchreq) = $dbh->prepare_cached(q{
		SELECT
			building_id,
			building_code,
			building_name,
			building_type,
			has_subunits,
			sites.site_name
		FROM buildings,sites
		WHERE building_type LIKE ?
		AND buildings.site_id = sites.site_id
		ORDER BY building_name
	});
	if (!defined($sth_fetchreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetchreq->execute(
		$typecode
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth_fetchreq->fetchall_arrayref;
	if ($sth_fetchreq->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $requests;
}


########################################################################
#
# Fetch list of sites:
#	FetchBuildingList
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchSiteList
{
	# Prepare and cache the select statement
	my ($sth_fetchreq) = $dbh->prepare_cached(q{
		SELECT
			site_id,
			site_name
		FROM sites
		ORDER BY site_name
	});
	if (!defined($sth_fetchreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetchreq->execute() ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($sites) = $sth_fetchreq->fetchall_arrayref;
	if ($sth_fetchreq->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $sites;
}

########################################################################
#
# Fetch list of (current/future) semesters
# Returns ref to array of arrays on success, zero on failure
# Silently ignores past semesters and tries to put most appropriate
# choice first
#
########################################################################
sub FetchSemesterList
{
	# Prepare and cache the select statement
	my ($sth_fetch) = $dbh->prepare_cached(q{
		SELECT
			sem_id,
			sem_start_date,
			sem_end_date,
			connection_fee,
			sem_fee,
			sem_description,
			sort_order
		FROM semesters
		WHERE sem_end_date > NOW()
		AND sort_order != 0
		ORDER BY sort_order, sem_end_date, sem_fee
	});
	if (!defined($sth_fetch)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetch->execute() ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($semesters) = $sth_fetch->fetchall_arrayref;
	if ($sth_fetch->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $semesters;
}

########################################################################
#
# Fetch list of semesters that are valid extensions
# for an existing connection whose end-date is passed as param 1.
# These are any semesters starting within `param 2' days of `param 1'.
# Returns ref to array of arrays on success, zero on failure
# Silently ignores past semesters and tries to put most appropriate
# choice first
#
########################################################################
sub FetchRenewalSemesterList
{
	my (
		$current_end_date,
		$days_tolerance
	) = @_;

	# Prepare and cache the select statement
	# We want semesters that start within a few days of the end
	# of the current connection period
	my ($sth_fetch) = $dbh->prepare_cached(q{
		SELECT
			sem_id,
			sem_start_date,
			sem_end_date,
			connection_fee,
			sem_fee,
			sem_description,
			sort_order
		FROM semesters
		WHERE sem_end_date > NOW()
		AND sort_order != 0
		AND ABS( TO_DAYS(sem_start_date) - TO_DAYS(?) ) < ?
		ORDER BY sort_order, sem_end_date, sem_fee
	});
	if (!defined($sth_fetch)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetch->execute(
			$current_end_date,
			$days_tolerance
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($semesters) = $sth_fetch->fetchall_arrayref;
	if ($sth_fetch->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $semesters;
}

########################################################################
#
# Fetch details of one semester
# Returns array on success, zero on failure
#
########################################################################
sub FetchSemester
{
	my ($sem_id) = @_;

	# Prepare and cache the select statement
	my ($sth) = $dbh->prepare_cached(q{
		SELECT
			sem_id,
			sem_start_date,
			sem_end_date,
			connection_fee,
			sem_fee,
			sem_description
		FROM semesters
		WHERE semesters.sem_id = ?
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth->execute( $sem_id )) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# There should only be one set of semester details under this ID
	if (defined($requests->[1]) || !defined($requests->[0])) {
		$ConnDB::errstr =
			"Internal error: Wrong number of semester details for ID $sem_id";
		return 0;
	}

	# We want to return an array, not a reference:
	return @{$requests->[0]};
}

########################################################################
#
# Fetch list of service requests.
# These can be limited by username and by request-id
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchRequests
{
	my (
		$username,$req_id
	) = @_;

	# Prepare and cache the select statement
	my ($sth_fetchreq) = $dbh->prepare_cached(q{
		SELECT
			service_request_id,
			username,
			service_type,
			request_date,
			charge,
			amount_paid,
			billing_details_id,
			comment,
			paid_date
		FROM service_request
		WHERE username LIKE ?
		AND service_request_id LIKE ?
		ORDER BY service_request_id
	});
	if (!defined($sth_fetchreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetchreq->execute(
		$username,
		$req_id
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth_fetchreq->fetchall_arrayref;
	if ($sth_fetchreq->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $requests;
}

########################################################################
#
# HallsRequestExists( $user, $start_date )
# - check for another service request that might indicate a duplicate
# Returns positive if duplicate exists, zero if not.
# Errors are returned as -1.
#
########################################################################
sub HallsRequestExists {
	my ($user, $start_date) = @_;

	my ($sth) = $dbh->prepare_cached(q{
		SELECT COUNT(*)
		FROM service_request,service_request_item
		WHERE service_request.username = ?
		AND service_request.service_request_id = service_request_item.service_request_id
		AND service_request_item.service_state_id IN (1,2,3,7)
		AND service_request_item.start_date = ?
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return -1;
	}

	# How many are there?
	my ($existing_requests) = $dbh->selectrow_array(
			$sth,undef,($user, $start_date));
	if (!defined($existing_requests)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return -1;
	}

	return $existing_requests;
}

########################################################################
#
# Fetch list of service request items relating to a given request number
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchRequestItemsByRequestID
{
	my (
		$req_num
	) = @_;

	# Prepare and cache the select statement
	my ($sth_fetchreq) = $dbh->prepare_cached(q{
		SELECT DISTINCT
			service_request_item.service_request_id,
			buildings.building_name,
			service_request_item.block_number,
			service_request_item.room_number,
			service_request_item.outlet_number,
			service_request_item.start_date,
			service_request_item.end_date,
			service_states.service_state_description,
			service_request_item.comment,
			buildings.building_name,	# This is a cheat:
							# we really want the building
							# name from the connection,
							# not the one from the request..
			connection.block_number,
			connection.room_number,
			connection.outlet_number,
			connection.device_id,
			devices.device_name,
			connection.device_port,
			connection.date_connected,
			connection.date_disconnected,
			connection.connected_by,
			connection.disconnected_by,
			service_request_item.service_request_item_id,
			service_request_item.service_state_id
		FROM service_request_item,buildings,service_states
		LEFT JOIN connection
			ON connection.service_request_item_id=service_request_item.service_request_item_id
		LEFT JOIN devices ON connection.device_id=devices.device_id
		WHERE service_request_id = ?
		AND service_request_item.service_state_id=service_states.service_state_id
		AND service_request_item.building_id=buildings.building_id
	});
	if (!defined($sth_fetchreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetchreq->execute(
		$req_num
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth_fetchreq->fetchall_arrayref;
	if ($sth_fetchreq->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $requests;
}


########################################################################
#
# Fetch a service request item
# Returns array on success, zero on failure
#
########################################################################
sub FetchRequestItem
{
	my (
		$req_item_num
	) = @_;

	# Prepare and cache the select statement
	my ($sth_fetchreq) = $dbh->prepare_cached(q{
		SELECT DISTINCT
			service_request_item.service_request_id,
			service_request_item.building_id,
			building_name,
			block_number,
			room_number,
			outlet_number,
			port_type,
			start_date,
			end_date,
			service_request_item.service_state_id,
			service_state_description,
			service_request_item.comment,
			service_request.username
		FROM service_request_item,buildings,service_states,service_request
		WHERE service_request_item_id = ?
		AND service_request_item.building_id=buildings.building_id
		AND service_request_item.service_state_id=service_states.service_state_id
		AND service_request_item.service_request_id=service_request.service_request_id
	});
	if (!defined($sth_fetchreq)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetchreq->execute(
		$req_item_num
	) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth_fetchreq->fetchall_arrayref;
	if ($sth_fetchreq->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}


	# There should only be one set of item details under this ID
	if (defined($requests->[1]) || !defined($requests->[0])) {
		$ConnDB::errstr =
			"Internal error: wrong number of service_request_items for ID $req_item_num";
		return 0;
	}

	# We want to return an array, not a reference:
	return @{$requests->[0]};
}


########################################################################
#
# Fetch a set of billing details
# Returns array on success, zero on failure
#
########################################################################
sub FetchBillingDetails
{
	my ($bill_id) = @_;

	# Prepare and cache the select statement
	my ($sth) = $dbh->prepare_cached(q{
		SELECT
			billing_details.card_account_name,
			billing_details.card_number,
			billing_details.card_expiry,
			billing_details.card_issue,
			billing_details.internal_charge_code
		FROM billing_details
		WHERE billing_details.billing_details_id = ?
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth->execute( $bill_id )) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# There should only be one set of billing details under this ID
	if (defined($requests->[1]) || !defined($requests->[0])) {
		$ConnDB::errstr =
			sprintf "Wrong number of billing details for ID $bill_id";
		return 0;
	}

	# We want to return an array, not a reference:
	return @{$requests->[0]};
}


########################################################################
#
# Fetch list of credit-card payments that need processing
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchPendingCardPayments
{
	# Prepare and cache the select statement
	my ($sth) = $dbh->prepare_cached(q{
		SELECT DISTINCT
			service_request.service_request_id,
			service_request.username,
			user.fullname,
			service_request.request_date,
			service_request.charge,
			service_request.amount_paid,
			service_request.billing_details_id,
			service_request.comment,
			billing_details.card_account_name,
			billing_details.card_number,
			billing_details.card_expiry,
			billing_details.card_issue
		FROM service_request,billing_details,user
		WHERE service_request.billing_details_id=billing_details.billing_details_id
		AND charge > amount_paid
		AND internal_charge_code = ""
		AND service_request.username = user.username
		ORDER BY username
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth->execute) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $requests;
}


########################################################################
#
# Fetch list of cash/cheque payments that need processing
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchPendingCashPayments
{
	# Prepare and cache the select statement
	my ($sth) = $dbh->prepare_cached(q{
		SELECT
			service_request.service_request_id,
			service_request.username,
			user.fullname,
			service_request.request_date,
			service_request.charge,
			service_request.amount_paid,
			service_request.billing_details_id,
			service_request.comment
		FROM service_request,user
		WHERE service_request.billing_details_id=0
		AND charge > amount_paid
		AND service_request.username = user.username
		ORDER BY username
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth->execute) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $requests;
}


########################################################################
#
# Fetch list of cash/cheque payments that are still awaited
# N days after service was requested.
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchLatePayments
{
	my ($days_ago) = @_;

	# Prepare and cache the select statement
	my ($sth) = $dbh->prepare_cached(q{
		SELECT
			service_request.service_request_id,
			service_request.username,
			user.fullname,
			service_request.request_date,
			service_request.charge,
			service_request.amount_paid,
			service_request.comment
		FROM service_request,user
		WHERE service_request.billing_details_id=0
		AND charge > amount_paid
		AND service_request.username = user.username
		AND TO_DAYS(NOW()) - TO_DAYS(service_request.request_date) > ?
		ORDER BY username
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth->execute($days_ago)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $requests;
}

########################################################################
#
# Fetch list of connections that will expire within the next N days
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchExpiringConnections
{
	my ($days_ahead) = @_;

	# Prepare and cache the select statement
	my ($sth) = $dbh->prepare_cached(q{
		SELECT
			service_request.username,
			user.fullname,
			service_request_item.end_date,
			buildings.building_name,
			connection.block_number,
			connection.room_number,
			connection.outlet_number
		FROM service_request,service_request_item,user,connection,buildings
		WHERE
			service_request_item.service_state_id = 3
			AND TO_DAYS(service_request_item.end_date) - TO_DAYS(NOW()) < ?
			AND service_request.username = user.username
			AND service_request.service_request_id=service_request_item.service_request_id
			AND connection.service_request_item_id=service_request_item.service_request_item_id
			AND connection.building_id=buildings.building_id
		ORDER BY username
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth->execute($days_ahead)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($expiring_list) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $expiring_list;
}


########################################################################
#
# Fetch list of connections current for a building
# Select by building_id, and optionally by block number (use '%' to get all)
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchConnectionList
{
	my (
		$building_id,
		$block_spec
	) = @_;

	# Prepare and cache the select statement
	my ($sth) = $dbh->prepare_cached(qq{
		SELECT DISTINCT
			connection.connection_id,
			buildings.building_id,
			buildings.building_name,
			connection.block_number,
			connection.room_number,
			connection.outlet_number,
			connection.device_id,
			devices.device_name,
			connection.device_port,
			connection.service_request_item_id,
			connection.date_connected,
			connection.connected_by,
			service_request.username
		FROM connection, buildings, devices, service_request_item, service_request
		WHERE connection.building_id = buildings.building_id
		AND connection.device_id = devices.device_id
		AND connection.date_disconnected = '0000-00-00 00:00:00'
		AND connection.building_id = ?
		AND connection.block_number LIKE ?
		AND connection.service_request_item_id = service_request_item.service_request_item_id
		AND service_request_item.service_request_id = service_request.service_request_id
		ORDER BY connection.block_number, connection.outlet_number
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth->execute(
		$building_id,
		$block_spec
	)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $requests;
}


########################################################################
#
# Fetch list of connections and disconnections to be made
# Select by campus, building (use '%' to get all)
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchPendingConnections
{
	my (
		$campus_spec,
		$building_spec
	) = @_;

	# Prepare and cache the select statement
	my ($sth) = $dbh->prepare_cached(qq{
		SELECT DISTINCT
			service_request_item.service_request_item_id,
			service_request.service_request_id,
			service_request.username,
			user.fullname,
			service_request.request_date,
			service_request.comment,
			service_request_item.building_id,
			buildings.building_name,
			sites.site_name,
			service_request_item.service_state_id,
			service_request_item.block_number,
			service_request_item.room_number,
			service_request_item.outlet_number,
			service_request_item.start_date,
			service_request_item.end_date,
			service_request_item.comment
		FROM service_request,user,service_request_item,buildings,sites
		WHERE service_request.service_request_id=service_request_item.service_request_id
		AND service_request.username=user.username
		AND service_request_item.building_id=buildings.building_id
		AND sites.site_id = buildings.site_id
		# State 2 is "awaiting connection"
		# State 4 is "awaiting disconnection"
		AND service_request_item.service_state_id IN (2,4)
		AND buildings.site_id LIKE ?
		AND buildings.building_id LIKE ?
		# Give 7 days warning of requested connections
		AND TO_DAYS(service_request_item.start_date) - TO_DAYS(NOW()) < $CONST_days_warning_of_connections
		ORDER BY buildings.site_id,buildings.building_name,
			service_request_item.block_number,
			service_request_item.room_number,
			service_request_item.outlet_number
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth->execute(
		$campus_spec,
		$building_spec
	)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($requests) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $requests;
}



########################################################################
#
# Process a payment for a pre-paid (i.e. Halls) connection
# Returns non-zero on success, zero on failure
#
########################################################################
sub ProcessPrePayment
{
	my (
		$service_request,
		$amount_paid
	) = @_;

	# Get details of this service_request
	my ($sth) = $dbh->prepare_cached(q{
		SELECT
			service_request.charge-service_request.amount_paid
		FROM service_request
		WHERE service_request.service_request_id=?
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Find out how much is due
	my ($amount_due) = $dbh->selectrow_array($sth,undef,($service_request));
	if (!defined($amount_due)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Does amount paid match amount due?
	if ($amount_due != $amount_paid) {
		$ConnDB::errstr = sprintf "Wrong payment amount: %5.2f due but %5.2f paid",
			$amount_due, $amount_paid;
		return 0;
	}

	# We need at least one item in "awaiting payment" state?
	my ($sth_numwaiting) = $dbh->prepare_cached(q{
		SELECT
			COUNT(*)
		FROM service_request_item
		WHERE service_request_item.service_state_id IN (1,7)
		AND service_request_id = ?
	});
	if (!defined($sth_numwaiting)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# How many are there?
	my ($number_waiting) = $dbh->selectrow_array($sth_numwaiting,undef,($service_request));
	if (!defined($number_waiting)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Some waiting to be paid?
	if ($number_waiting < 1) {
		$ConnDB::errstr = "No items are awaiting payment! (This should not happen)";
		return 0;
	}

	# OK - we can accept the payment and release the work to be done.

	# Release the work first to avoid upsetting the user if the second update fails!
	# First deal with any new connections - they need to become 'awaiting connection'
	my ($sth_setstate) = $dbh->prepare_cached(q{
		UPDATE service_request_item
		SET service_state_id = 2
		WHERE service_request_item.service_state_id = 1
		AND service_request_id = ?
	});
	if (!defined($sth_setstate)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	if (!$sth_setstate->execute($service_request) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Now deal with renewals - they revert to `connected' and the end_date gets updated
	$sth_setstate = $dbh->prepare_cached(q{
		UPDATE service_request_item
		SET service_state_id = 3,
			end_date = renew_to_date
		WHERE service_request_item.service_state_id = 7
		AND service_request_id = ?
	});
	if (!defined($sth_setstate)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	if (!$sth_setstate->execute($service_request) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Now record the payment
	my ($sth_setpaid) = $dbh->prepare_cached(q{
		UPDATE service_request
		SET
			amount_paid = amount_paid + ?,
			paid_date = NOW()
		WHERE service_request_id = ?
	});
	if (!defined($sth_setpaid)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	if (!$sth_setpaid->execute($amount_paid,$service_request) ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Find the billing_details_id
	my ($sth_billdet) = $dbh->prepare_cached(q{
		SELECT
			billing_details_id
		FROM service_request
		WHERE service_request.service_request_id=?
	});
	if (!defined($sth_billdet)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the ID
	my ($billing_details_id) = $dbh->selectrow_array($sth_billdet,undef,($service_request));
	if (!defined($billing_details_id)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	if (!DeleteCardDetails($billing_details_id)) {
		# Error
		return 1;
	}

	Log( "notice", "OK ProcessPrePayment $service_request, $amount_paid" );
	return 1;
}

########################################################################
#
# DeleteCardDetails
#
# Wipes out credit/debit card info by billing_details_id
# Returns 0 for fail, 1 for success
#
########################################################################
sub DeleteCardDetails {
	my ($billing_details_id) = @_;

	if ($billing_details_id > 0) {
		# Now wipe out any credit/debit card details that were
		# associated with this request
		my ($sth_wipe_card_details) = $dbh->prepare_cached(q{
			UPDATE billing_details
			SET card_account_name = "CARD DETAILS DELETED",
				card_number = "",
				card_expiry = "",
				card_issue = ""
			WHERE billing_details_id = ?
			# AND card_number LIKE "_%"
			# That finds non-empty card_number fields
		});
		if (!defined($sth_wipe_card_details)) {
			# Failed...
			$ConnDB::errstr = $DBI::errstr;
			return 0;
		}

		if (!$sth_wipe_card_details->execute($billing_details_id) ) {
			# Failed...
			$ConnDB::errstr = $DBI::errstr;
			return 0;
		}

		Log( "notice", "OK DeleteCardDetails $billing_details_id" );
	}

	# OK
	return 1;
}

########################################################################
#
# MarkExpiredConnections
#
# Set connections to be removed if past their end-date
#
########################################################################

sub MarkExpiredConnections {
	# Find expired connections and mark for disconnection
	my ($sth_find_expired) = $dbh->prepare_cached(q{
		SELECT service_request_item_id
		FROM service_request_item
		WHERE end_date < NOW()
		AND service_state_id IN (1,2,3,7)
	});
	if (!defined($sth_find_expired)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}
	if (!$sth_find_expired->execute()) {
		# Failed to update...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($item_list) = $sth_find_expired->fetchall_arrayref;
	if ($sth_find_expired->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Walk through the array and cancel each connection
	my ($item);
	for ($item=0;defined($item_list->[$item]);$item++) {
		my ($thisitem) = $item_list->[$item][0];
		print "Cancelling item $thisitem\n";
		CancelRequestItem( $thisitem, "system", "End of booked connection period", 1);
	}

	Log( "notice", "OK MarkExpiredConnections" );

	# OK
	return 1;
}

########################################################################
#
# CancelRequestItem( $item_id, $loggedinuser, $reason, $no_check )
#
# Expects params to be sanity-checked
#
# If no_check is non-zero then permissions are not checked.
#
# Returns:
#	0 on success
#	1 for permission denied
#	2 for unable to cancel
#	3 for internal error
#
########################################################################
sub CancelRequestItem {
	my ($item_id, $loggedinuser, $reason, $no_check ) = @_;

	# We need some info about this item
	my ($sth) = $dbh->prepare_cached(q{
		SELECT
			service_request_item.service_request_id,
			service_request_item.service_state_id,
			service_request.username,
			service_request.billing_details_id,
			service_request.charge,
			service_request_item.charge
		FROM service_request_item,service_request
		WHERE service_request_item.service_request_id=service_request.service_request_id
		AND service_request_item.service_request_item_id = ?
	});
	if (!defined($sth)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 3;
	}

	# Run the select
	if (!$sth->execute( $item_id )) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 3;
	}

	# Fetch the results
	my ($item_details) = $sth->fetchall_arrayref;
	if ($sth->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 3;
	}

	# There should only be one set of item details under this ID
	if (defined($item_details->[1]) || !defined($item_details->[0])) {
		$ConnDB::errstr = "Wrong number of item details for ID $item_id";
		return 3;
	}

	# Extract the data
	my (@arr) = @{$item_details->[0]};
	my ($service_request_id, $service_state_id, $username,
		$billing_details_id, $sr_charge, $sri_charge) = @arr;

	# Is this person allowed to cancel this request?
	if ( !(		($loggedinuser eq $username) ||
			Permitted( $loggedinuser, "billing" ) ||
			Permitted( $loggedinuser, "connect" ) ||
			$no_check
	)) {
		#Permission denied
		return 1;
	}

	# Is the request cancellable anyway?
	if ( !($service_state_id =~ /^(1|2|3|7)$/) ) {
		# Can only cancel items in states 1,2,3,7
		return 2;
	}

	# What does the money look like?
	# PROBLEM: renewed connections result in charges changing after connection.
	# This causes problems with cancelling items.
	#if ($sr_charge < $sri_charge) {
	#	$ConnDB::errstr = "Cancelling that item would take the total bill negative";
	#	return 3;
	#}

	########################################
	# OK - looks possible so lets cancel it
	########################################

	my ($comment) = substr "$reason", 0,
				$CONST_max_comment_length;
	my ($new_state);
	if (($service_state_id == 1) || ($service_state_id ==2)) {
		# We go from any pre-connect state to cancelled
		$new_state = 6;
	}
	else {
		# Other states require disconnection
		$new_state = 4;
	}

	# PROBLEM: We should subtract the cost of this line-item from the total charge
	# but the problem with renewed connections messes it up.
	# For the moment we will make the simplifying assumption that each service_request
	# will only contain one service_request_item
	#
	my ($sth_sub) = $dbh->prepare_cached(q{
		UPDATE service_request
		SET
			charge = 0,
			amount_paid = 0
		WHERE service_request_id = ?
	});
	if (!defined($sth_sub)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 3;
	}
	if (!$sth_sub->execute( $service_request_id )) {
		# Failed to update...
		$ConnDB::errstr = $DBI::errstr;
		return 3;
	}

	# Update the item state
	my ($sth_item) = $dbh->prepare_cached(q{
		UPDATE service_request_item
		SET charge = 0, service_state_id = ?, comment = ?
		WHERE service_request_item_id = ?
	});
	if (!defined($sth_item)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 3;
	}
	if (!$sth_item->execute( $new_state, $comment, $item_id )) {
		# Failed to update...
		$ConnDB::errstr = $DBI::errstr;
		return 3;
	}



	# Wipe out card details if there are any
	if (!DeleteCardDetails($billing_details_id)) {
		# Error
		return 3;
	}

	Log( "notice", "OK CancelRequestItem $item_id, $loggedinuser, $reason" );

	return 0;
}


########################################################################
#
# Fetch list of users with special privileges
# Returns ref to array of arrays on success, zero on failure
# Data is sorted by username
#
########################################################################
sub FetchPermissionsList
{
	# Prepare and cache the select statement
	my ($sth_fetch) = $dbh->prepare_cached(q{
		SELECT
			permissions.username,
			permissions.priv,
			user.fullname
		FROM permissions
		LEFT JOIN user ON permissions.username = user.username
		ORDER BY username, priv
	});
	if (!defined($sth_fetch)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetch->execute() ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($items) = $sth_fetch->fetchall_arrayref;
	if ($sth_fetch->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $items;
}


########################################################################
#
# Fetch list of special privileges
# Returns ref to array of arrays on success, zero on failure
#
########################################################################
sub FetchPrivsList
{
	# Prepare and cache the select statement
	my ($sth_fetch) = $dbh->prepare_cached(q{
		SELECT
			priv,
			priv_description
		FROM privs
		ORDER BY priv
	});
	if (!defined($sth_fetch)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Run the select
	if (!$sth_fetch->execute() ) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Fetch the results
	my ($items) = $sth_fetch->fetchall_arrayref;
	if ($sth_fetch->err) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $items;
}

########################################################################
#
# Check whether a user has a particular permission
# Returns true if permission is granted, zero if not.
# Also returns zero for errors to avoid security risk.
#
########################################################################
sub Permitted {
	my ($username, $priv) = @_;

	my ($sth_get) = $dbh->prepare_cached(q{
		SELECT
			COUNT(*)
		FROM permissions
		WHERE username = ?
		AND priv = ?
	});
	if (!defined($sth_get)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# How many are there? (should be zero or one)
	my ($number) = $dbh->selectrow_array($sth_get,undef,($username, $priv));
	if (!defined($number)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	return $number;
}

########################################################################
#
# SetPermissions( $user, $has_admin, $has_billing, $has_connect );
#
# Returns positive if OK, zero if not
#
########################################################################
sub SetPermissions {
	my ($user, $has_admin, $has_billing, $has_connect) = @_;

	# This works by removing all existing permissions for the user
	# then adding back the required ones.
	# Note that there is a small window where the user has no
	# permissions so any failure or other access in that time
	# will be a problem

	# Delete any existing permissions
	my ($sth_del) = $dbh->prepare_cached(q{
		DELETE FROM permissions
		WHERE username = ?
	});
	if (!defined($sth_del)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}
	if (!$sth_del->execute( $user )) {
		# Failed to update...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# Add new permissions
	my ($sth_ins) = $dbh->prepare_cached(q{
		INSERT INTO permissions
		VALUES ( ?, ? )
	});
	if (!defined($sth_ins)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	if ($has_admin) {
		if (!$sth_ins->execute( $user, "admin" )) {
			# Failed to update...
			$ConnDB::errstr = $DBI::errstr;
			return 0;
		}
	}

	if ($has_billing) {
		if (!$sth_ins->execute( $user, "billing" )) {
			# Failed to update...
			$ConnDB::errstr = $DBI::errstr;
			return 0;
		}
	}

	if ($has_connect) {
		if (!$sth_ins->execute( $user, "connect" )) {
			# Failed to update...
			$ConnDB::errstr = $DBI::errstr;
			return 0;
		}
	}

	Log( "notice", "OK SetPermissions $user, $has_admin, $has_billing, $has_connect" );

	return 1;
}

########################################################################
#
# Check to see whether a given outlet is already known
#
########################################################################
sub KnownOutletNumber {
	my ($building_id, $block, $outlet_number) = @_;


	# We need to match exactly one item
	# (there cannot be more as this is the primary key!)
	my ($sth_get) = $dbh->prepare_cached(q{
		SELECT
			COUNT(*)
		FROM outlets
		WHERE building_id = ?
		AND block_number = ?
		AND outlet_number = ?
	});
	if (!defined($sth_get)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# How many are there?
	my ($number) = $dbh->selectrow_array($sth_get,undef,
			($building_id, $block, $outlet_number));
	if (!defined($number)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# No matching values found
	if ($number < 1) {
		return 0;
	}

	# OK - value is in table
	return 1;
}

########################################################################
#
# LoadOutletData( $building_id, @outlets )
#
########################################################################

sub LoadOutletData {
	my ( $building_id, $OutletArray, $IgnoreDuplicates, $delete_flag ) = @_;

	# Prepare and cache the insert statement
	my ($sth_addoutlet) = $dbh->prepare_cached(q{
		INSERT INTO outlets (
			building_id,
			block_number,
			room_number,
			outlet_number,
			comment
		)
		VALUES (?,?,?,?,?)
	});
	if (!defined($sth_addoutlet)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		Log( "error", "LoadOutletData prepare problem: $DBI::errstr" );
		return [-1,0,0];
	}

	# Should we clear out old data first?
	my ($ndeleted) = 0;
	if ($delete_flag) {
		my ($rv) = $dbh->do( q{
			DELETE FROM outlets
			WHERE building_id = ?
		}, 0, ($building_id));

		if (!defined($rv)) {
			$ConnDB::errstr = $DBI::errstr;
			Log( "error", "LoadOutletData delete problem: $DBI::errstr" );
			return [-1,0,0];
		}
		$ndeleted = $rv;
	}

	# Work through the supplied array adding data to the table
	my ($count) = 0;
	my ($loaded) = 0;
	my ($duplicates) = 0;
	while (defined($OutletArray->[$count][2])) {
		if ($IgnoreDuplicates &&
			KnownOutletNumber(
				$building_id,
				$OutletArray->[$count][0],
				$OutletArray->[$count][2] )) {
			$duplicates++;
			$count++;
			next;
		}

		if (!$sth_addoutlet->execute(
			$building_id,
			$OutletArray->[$count][0],
			$OutletArray->[$count][1],
			$OutletArray->[$count][2],
			$OutletArray->[$count][3]
			))
		{
			# Failed...
			$ConnDB::errstr = $DBI::errstr;
			Log( "error", "LoadOutletData execute problem: $DBI::errstr" );
			return [-1,0,0];
		}
		$loaded++;
		$count++;
	}

	Log( "notice", "LoadOutletData loaded $loaded items and found $duplicates duplicates for building $building_id" );
	return [$loaded, $duplicates, $ndeleted];
}

########################################################################
#
# GetCounts
#
# Return string of numbers indicating how many request items are
# in each of the possible service states
#
########################################################################
sub GetCountForState {
	my ($state_number) = @_;


	my ($sth_get) = $dbh->prepare_cached(q{
		SELECT
			COUNT(*)
		FROM service_request_item
		WHERE service_state_id = ?
	});
	if (!defined($sth_get)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# How many are there?
	my ($number) = $dbh->selectrow_array($sth_get,undef,
			($state_number));
	if (!defined($number)) {
		# Failed...
		$ConnDB::errstr = $DBI::errstr;
		return 0;
	}

	# No matching values found
	if ($number < 1) {
		return 0;
	}

	# OK - return number found
	return $number;
}

sub GetCounts {
	my ($this_state) = 0;

	my ($result) = "";

	while ($this_state <= $CONST_max_state_num) {
		my ($number) = GetCountForState( $this_state );
		$result .= "$number ";
		$this_state++;
	}

	# Remove trailing blank
	$result =~ s/ $//;

	return $result;
}

########################################################################
#
# Mark end of ConnDB.pm with true return
#
########################################################################

1;

########################################################################
########################################################################
########################################################################
